#  !/usr/bin/env python
#   -*- coding: utf-8 -*-
#
#  __init__.py
#
#  Copyright (c) 2019 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  assay, Atom, Bond, Compound, Constants, Errors, Lookup, Substance and
#  Utils based on PubChemPy by Matt Swain <m.swain@me.com>
#  Available under the MIT License
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

__author__ = "Dominic Davis-Foster"
__copyright__ = "2019 Dominic Davis-Foster"

__license__ = "LGPL"
__version__ = "0.2.0"
__email__ = "dominic@davis-foster.co.uk"


import requests_cache

# Setup Cache and keep for ~ a month
requests_cache.install_cache("chemistry_tools_cache", expire_after=2500000)


from . import assay
from . import atom
from . import bond
from . import compound
from . import constants
from . import errors
from . import lookup
from . import property_format
from . import spectrum_similarity
from . import substance
from . import toxnet
from . import utils


def clear_cache():
	requests_cache.clear()


__all__ = [
		"spectrum_similarity",
		"assay",
		"atom",
		"bond",
		"compound",
		"constants",
		"errors",
		"property_format",
		"substance",
		"toxnet",
		"utils",
		]

if __name__ == '__main__':
	print(__version__)




#!/usr/bin/env python
#
#
#  atom.py
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

from .constants import ELEMENTS
from .errors import deprecated


class Atom:
	"""Class to represent an atom in a :class:`~pubchempy.Compound`."""
	
	def __init__(self, aid, number, x=None, y=None, z=None, charge=0):
		"""Initialize with an atom ID, atomic number, coordinates and optional change.

		:param int aid: Atom ID
		:param int number: Atomic number
		:param float x: X coordinate.
		:param float y: Y coordinate.
		:param float z: (optional) Z coordinate.
		:param int charge: (optional) Formal charge on atom.
		"""
		
		self.aid = aid
		# The atom ID within the owning Compound.
		self.number = number
		# The atomic number for this atom.
		self.x = x
		# The x coordinate for this atom.
		self.y = y
		# The y coordinate for this atom.
		self.z = z
		# The z coordinate for this atom. Will be ``None`` in 2D Compound records.
		self.charge = charge
		# The formal charge on this atom.
	
	def __repr__(self):
		return 'Atom({}, {})'.format(self.aid, self.element)
	
	def __eq__(self, other):
		return (
				isinstance(other, type(self)) and
				self.aid == other.aid and
				self.element == other.element and
				self.x == other.x and
				self.y == other.y and
				self.z == other.z and
				self.charge == other.charge
				)
	
	@deprecated('Dictionary style access to Atom attributes is deprecated')
	def __getitem__(self, prop):
		"""Allow dict-style access to attributes to ease transition from when atoms were dicts."""
		if prop in {'element', 'x', 'y', 'z', 'charge'}:
			return getattr(self, prop)
		raise KeyError(prop)
	
	@deprecated('Dictionary style access to Atom attributes is deprecated')
	def __setitem__(self, prop, val):
		"""Allow dict-style setting of attributes to ease transition from when atoms were dicts."""
		setattr(self, prop, val)
	
	@deprecated('Dictionary style access to Atom attributes is deprecated')
	def __contains__(self, prop):
		"""Allow dict-style checking of attributes to ease transition from when atoms were dicts."""
		if prop in {'element', 'x', 'y', 'z', 'charge'}:
			return getattr(self, prop) is not None
		return False
	
	@property
	def element(self):
		"""The element symbol for this atom."""
		return ELEMENTS.get(self.number, None)
	
	def to_dict(self):
		"""Return a dictionary containing Atom data."""
		data = {'aid': self.aid, 'number': self.number, 'element': self.element}
		for coord in {'x', 'y', 'z'}:
			if getattr(self, coord) is not None:
				data[coord] = getattr(self, coord)
		if self.charge != 0:
			data['charge'] = self.charge
		return data
	
	def set_coordinates(self, x, y, z=None):
		"""Set all coordinate dimensions at once."""
		self.x = x
		self.y = y
		self.z = z
	
	@property
	def coordinate_type(self):
		"""Whether this atom has 2D or 3D coordinates."""
		return '2d' if self.z is None else '3d'



#  !/usr/bin/env python
#
#  utils.py
"""Various tools"""
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
import functools
import json
import os
import time
from decimal import Decimal, DecimalException
# from urllib.request import urlopen
# from urllib.error import HTTPError
from urllib.parse import quote, urlencode

# 3rd party
import requests
# from requests.exceptions import HTTPError

# this package
from .constants import API_BASE, log, PROPERTY_MAP, text_types
from .errors import NotFoundError, PubChemHTTPError


def get_full_json(cid):
	json_file = requests.get(f"https://pubchem.ncbi.nlm.nih.gov/rest/pug_view/data/compound/{cid}/JSON/")
	return json_file.json()


def get_json(identifier, namespace='cid', domain='compound', operation=None, searchtype=None, **kwargs):
	"""
	Request wrapper that automatically parses JSON response and suppresses NotFoundError.
	"""
	
	try:
		return json.loads(get(identifier, namespace, domain, operation, 'JSON', searchtype, **kwargs).decode())
	except NotFoundError as e:
		log.info(e)
		return None


def get(identifier, namespace='cid', domain='compound', operation=None, output='JSON', searchtype=None, **kwargs):
	"""
	Request wrapper that automatically handles async requests.
	"""
	
	if (searchtype and searchtype != 'xref') or namespace in ['formula']:
		# response = request(identifier, namespace, domain, None, 'JSON', searchtype, **kwargs).read()
		response = request(identifier, namespace, domain, None, 'JSON', searchtype, **kwargs).content
		status = json.loads(response.decode())
		if 'Waiting' in status and 'ListKey' in status['Waiting']:
			identifier = status['Waiting']['ListKey']
			namespace = 'listkey'
			while 'Waiting' in status and 'ListKey' in status['Waiting']:
				time.sleep(2)
				# response = request(identifier, namespace, domain, operation, 'JSON', **kwargs).read()
				response = request(identifier, namespace, domain, operation, 'JSON', **kwargs).content
				status = json.loads(response.decode())
			if not output == 'JSON':
				# response = request(identifier, namespace, domain, operation, output, searchtype, **kwargs).read()
				response = request(identifier, namespace, domain, operation, output, searchtype, **kwargs).content
	else:
		# response = request(identifier, namespace, domain, operation, output, searchtype, **kwargs).read()
		response = request(identifier, namespace, domain, operation, output, searchtype, **kwargs).content
	return response


def request(identifier, namespace='cid', domain='compound', operation=None, output='JSON', searchtype=None, **kwargs):
	"""
	Construct API request from parameters and return the response.

	Full specification at http://pubchem.ncbi.nlm.nih.gov/pug_rest/PUG_REST.html
	"""
	
	if not identifier:
		raise ValueError('identifier/cid cannot be None')
	# If identifier is a list, join with commas into string
	if isinstance(identifier, int):
		identifier = str(identifier)
	if not isinstance(identifier, text_types):
		identifier = ','.join(str(x) for x in identifier)
	# Filter None values from kwargs
	kwargs = {k: v for k, v in kwargs.items() if v is not None}
	# Build API URL
	urlid, postdata = None, None
	if namespace == 'sourceid':
		identifier = identifier.replace('/', '.')
	if namespace in ['listkey', 'formula', 'sourceid'] \
			or searchtype == 'xref' \
			or (searchtype and namespace == 'cid') or domain == 'sources':
		urlid = quote(identifier.encode('utf8'))
	else:
		postdata = urlencode([(namespace, identifier)]).encode('utf8')
	comps = filter(None, [API_BASE, domain, searchtype, namespace, urlid, operation, output])
	apiurl = '/'.join(comps)
	if kwargs:
		apiurl += '?%s' % urlencode(kwargs)
	# Make request
	# try:
	log.debug('Request URL: %s', apiurl)
	log.debug('Request data: %s', postdata)
	# response = urlopen(apiurl, postdata)
	response = requests.get(apiurl, postdata)
	if response.status_code in ERROR_CODES:
		raise PubChemHTTPError(response)
		# print(f"#{response}")
	return response
	# except HTTPError as e:
	# 	raise PubChemHTTPError(e)

ERROR_CODES = [400, 404, 405, 504, 501, 500]


def get_sdf(identifier, namespace='cid', domain='compound', operation=None, searchtype=None, **kwargs):
	"""
	Request wrapper that automatically parses SDF response and suppresses NotFoundError.
	"""
	
	try:
		return get(identifier, namespace, domain, operation, 'SDF', searchtype, **kwargs).decode()
	except NotFoundError as e:
		log.info(e)
		return None


def get_properties(properties, identifier, namespace='cid', searchtype=None, as_dataframe=False, **kwargs):
	"""
	Retrieve the specified properties from PubChem.

	:param identifier: The compound, substance or assay identifier to use as a search query.
	:param namespace: (optional) The identifier type.
	:param searchtype: (optional) The advanced search type, one of substructure, superstructure or similarity.
	:param as_dataframe: (optional) Automatically extract the properties into a pandas :class:`~pandas.DataFrame`.
	"""
	
	if isinstance(properties, text_types):
		properties = properties.split(',')
	properties = ','.join([PROPERTY_MAP.get(p, p) for p in properties])
	properties = 'property/%s' % properties
	results = get_json(identifier, namespace, 'compound', properties, searchtype=searchtype, **kwargs)
	results = results['PropertyTable']['Properties'] if results else []
	if as_dataframe:
		import pandas as pd
		return pd.DataFrame.from_records(results, index='CID')
	return results


def get_synonyms(identifier, namespace='cid', domain='compound', searchtype=None, **kwargs):
	results = get_json(identifier, namespace, domain, 'synonyms', searchtype=searchtype, **kwargs)
	return results['InformationList']['Information'] if results else []


def get_cids(identifier, namespace='name', domain='compound', searchtype=None, **kwargs):
	results = get_json(identifier, namespace, domain, 'cids', searchtype=searchtype, **kwargs)
	if not results:
		return []
	elif 'IdentifierList' in results:
		return results['IdentifierList']['CID']
	elif 'InformationList' in results:
		return results['InformationList']['Information']


def get_sids(identifier, namespace='cid', domain='compound', searchtype=None, **kwargs):
	results = get_json(identifier, namespace, domain, 'sids', searchtype=searchtype, **kwargs)
	if not results:
		return []
	elif 'IdentifierList' in results:
		return results['IdentifierList']['SID']
	elif 'InformationList' in results:
		return results['InformationList']['Information']


def get_aids(identifier, namespace='cid', domain='compound', searchtype=None, **kwargs):
	results = get_json(identifier, namespace, domain, 'aids', searchtype=searchtype, **kwargs)
	if not results:
		return []
	elif 'IdentifierList' in results:
		return results['IdentifierList']['AID']
	elif 'InformationList' in results:
		return results['InformationList']['Information']


def get_all_sources(domain='substance'):
	"""
	Return a list of all current depositors of substances or assays.
	"""
	
	results = json.loads(get(domain, None, 'sources').decode())
	return results['InformationList']['SourceName']


def download(
		outformat, path, identifier, namespace='cid', domain='compound',
		operation=None, searchtype=None, overwrite=False, **kwargs):
	"""
	Format can be  XML, ASNT/B, JSON, SDF, CSV, PNG, TXT.
	"""
	
	response = get(identifier, namespace, domain, operation, outformat, searchtype, **kwargs)
	if not overwrite and os.path.isfile(path):
		raise OSError("%s already exists. Use 'overwrite=True' to overwrite it." % path)
	with open(path, 'wb') as f:
		f.write(response)


def memoized_property(fget):
	"""
	Decorator to create memoized properties.

	Used to cache :class:`~pubchempy.Compound` and :class:`~pubchempy.Substance` properties that require an additional
	request.
	"""
	
	attr_name = '_{}'.format(fget.__name__)
	
	@functools.wraps(fget)
	def fget_memoized(self):
		if not hasattr(self, attr_name):
			setattr(self, attr_name, fget(self))
		return getattr(self, attr_name)
	
	return property(fget_memoized)


def _parse_prop(search, proplist):
	"""
	Extract property value from record using the given urn search filter.
	"""
	
	props = [i for i in proplist if all(item in i['urn'].items() for item in search.items())]
	if len(props) > 0:
		if search != {'implementation': 'E_SCREEN'}:  # True for "fingerprint", which isn't a number
			try:
				return Decimal(str(props[0]['value'][list(props[0]['value'].keys())[0]]))
			except DecimalException:
				return props[0]['value'][list(props[0]['value'].keys())[0]]
		else:
			return props[0]['value'][list(props[0]['value'].keys())[0]]


def format_string(stringwithmarkup):
	"""
	Convert a PubChem formatted string into an HTML formatted string
	"""
	
	string = list(stringwithmarkup["String"])
	try:
		markup_list = stringwithmarkup["Markup"]
	except KeyError:
		markup_list = []
	
	for markup in markup_list:
		style = None
		start = markup["Start"]
		end = markup["Length"] + start - 1
		if markup["Type"] == "Italics":
			style = "i"
		# handle Other formats
		
		if style is None:
			print(markup)
			continue
		
		string[start] = f"<{style}>{string[start]}"
		string[end] = f"{string[end]}</{style}>"
	
	string = ''.join(string)
	
	return string



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
#
# Copyright 2019 Dominic Davis-Foster
#
# Adapted from SpectrumSimilarity.R
# Part of OrgMassSpecR
# Copyright 2011-2017 Nathan Dodder <nathand@sccwrp.org>
# Available under the BSD 2-Clause License


__author__ = "Dominic Davis-Foster"
__copyright__ = "Copyright 2019 Dominic Davis-Foster"

__license__ = "LGPL"
__version__ = "0.1.0"
__email__ = "dominic@davis-foster.co.uk"


import numpy as np
import pandas as pd

def SpectrumSimilarity(spec_top, spec_bottom, t = 0.25, b = 10, top_label = None, 
                               bottom_label = None, xlim = (50, 1200), x_threshold = 0, print_alignment = False,
                               print_graphic = True, output_list = False):
	"""
	Parameters:
	spec_top (numpy.array): Array containing the experimental spectrum's peak list with the m/z values in the
		first column and corresponding intensities in the second
	spec_bottom	(numpy.array): Array containing the reference spectrum's peak list with the m/z values in the
		first column and corresponding intensities in the second
	t (float): numeric value specifying the tolerance used to align the m/z values of the two spectra.
	b (float): numeric value specifying the baseline threshold for peak identification. Expressed as a percent of the maximum intensity.
	top_label (str): string to label the top spectrum.
	bottom.label  (str): string to label the bottom spectrum.	
	xlim (tuple of ints): tuple of length 2, defining the beginning and ending values of the x-axis.
	x_threshold (float): numeric value specifying
	print_alignment (bool): whether the intensities should be printed
	print_graphic (bool):
	output_list (bool): whether the intensities should be returned
	"""
	if print_graphic:
		import matplotlib.pyplot as plt

	
	## format spectra and normalize intensitites
	top_tmp = pd.DataFrame(data = spec_top, columns = ["mz", "intensity"])
	top_tmp["normalized"] = top_tmp.apply(normalize, args=(max(top_tmp["intensity"]),),axis=1)
	top_tmp = top_tmp[top_tmp["mz"].between(xlim[0],xlim[1])]
	top_plot = top_tmp[["mz", "normalized"]].copy() # data frame for plotting spectrum
	top_plot.columns = ["mz", "intensity"]
	top = top_plot[top_plot["intensity"] >= b] # data frame for similarity score calculation

	bottom_tmp = pd.DataFrame(data = spec_bottom, columns = ["mz", "intensity"])
	bottom_tmp["normalized"] = bottom_tmp.apply(normalize, args=(max(bottom_tmp["intensity"]),),axis=1)
	bottom_tmp = bottom_tmp[bottom_tmp["mz"].between(xlim[0],xlim[1])]
	bottom_plot = bottom_tmp[["mz", "normalized"]].copy() # data frame for plotting spectrum
	bottom_plot.columns = ["mz", "intensity"]
	bottom = bottom_plot[bottom_plot["intensity"] >= b] # data frame for similarity score calculation
	
	
	## align the m/z axis of the two spectra, the bottom spectrum is used as the reference
	
	#Unimplemented R code
	"""
	  for(i in 1:nrow(bottom))
		top["mz"][bottom["mz"][i] >= top["mz"] - t & bottom["mz"][i] <= top["mz"] + t] = bottom["mz"][i]
		top[,1][bottom[,1][i] >= top[,1] - t & bottom[,1][i] <= top[,1] + t] <- bottom[,1][i]
	  alignment <- merge(top, bottom, by = 1, all = TRUE)
	  if(length(unique(alignment[,1])) != length(alignment[,1])) warning("the m/z tolerance is set too high")
	alignment[,c(2,3)][is.na(alignment[,c(2,3)])] <- 0   # convert NAs to zero (R-Help, Sept. 15, 2004, John Fox)
	names(alignment) <- c("mz", "intensity.top", "intensity.bottom")
	"""
	alignment = pd.merge(top, bottom, on="mz", how="outer")
	alignment.fillna(value=0, inplace=True) # Convert NaN to 0
	alignment.columns=["mz", "intensity_top", "intensity_bottom"]
	if print_alignment:
		with pd.option_context('display.max_rows', None, 'display.max_columns', None):
			print(alignment)
	
	## similarity score calculation
	
	if x_threshold < 0:
		print("Error: x_threshold argument must be zero or a positive number")
		return 1
	
	
	#Unimplemented R code
	"""
	alignment <- alignment[alignment[,1] >= x.threshold, ]
	"""
	
	u = np.array(alignment.iloc[:,1])
	v = np.array(alignment.iloc[:,2])
	
	similarity_score = np.dot(u,v) / (np.sqrt(np.sum(np.square(u))) * np.sqrt(np.sum(np.square(v))))
	
	
	# Reverse Match
	reverse_alignment = pd.merge(top, bottom, on="mz", how="right")
	reverse_alignment = reverse_alignment.dropna() #Remove rows containing NaN
	reverse_alignment.columns=["mz", "intensity_top", "intensity_bottom"]
	u = np.array(reverse_alignment.iloc[:,1])
	v = np.array(reverse_alignment.iloc[:,2])
	
	reverse_similarity_score = np.dot(u,v) / (np.sqrt(np.sum(np.square(u))) * np.sqrt(np.sum(np.square(v))))
	
	## generate plot
	
	if print_graphic:
		fig, ax = plt.subplots()
		#fig.scatter(top_plot["mz"],top_plot["intensity"], s=0)
		ax.vlines(top_plot["mz"], 0, top_plot["intensity"], color="blue")
		ax.vlines(top_plot["mz"], 0, -top_plot["intensity"], color="red")
		ax.set_ylim(-125,125)
		ax.set_xlim(xlim[0], xlim[1])
		ax.axhline(color="black", linewidth=0.5)
		ax.set_ylabel("Intensity (%)")
		ax.set_xlabel("m/z", style="italic", family="serif")
		
		h_centre = xlim[0]+(xlim[1]-xlim[0])//2
		
		ax.text(h_centre, 110, top_label, horizontalalignment = "center", verticalalignment = "center")
		ax.text(h_centre, -110, bottom_label,horizontalalignment = "center", verticalalignment = "center")
		plt.show()
		
	#Unimplemented R code	
	"""   
		ticks <- c(-100, -50, 0, 50, 100)
		plot.window(xlim = c(0, 20), ylim = c(-10, 10))
	  
	  
	  if(output.list == TRUE) {
		
		# Grid graphics head to tail plot
		
		headTailPlot <- function() {
		  
		  pushViewport(plotViewport(c(5, 5, 2, 2)))
		  pushViewport(dataViewport(xscale = xlim, yscale = c(-125, 125)))
		  
		  grid.rect()
		  tmp <- pretty(xlim)
		  xlabels <- tmp[tmp >= xlim[1] & tmp <= xlim[2]]
		  grid.xaxis(at = xlabels)
		  grid.yaxis(at = c(-100, -50, 0, 50, 100))
		  
		  grid.segments(top_plot$mz,
						top_plot$intensity,
						top_plot$mz,
						rep(0, length(top_plot$intensity)),
						default.units = "native",
						gp = gpar(lwd = 0.75, col = "blue"))
		  
		  grid.segments(bottom_plot$mz,
						-bottom_plot$intensity,
						bottom_plot$mz,
						rep(0, length(bottom_plot$intensity)),
						default.units = "native",
						gp = gpar(lwd = 0.75, col = "red"))
		  
		  grid.abline(intercept = 0, slope = 0)
		  
		  grid.text("intensity (%)", x = unit(-3.5, "lines"), rot = 90)
		  grid.text("m/z", y = unit(-3.5, "lines"))
		  
		  popViewport(1)
		  pushViewport(dataViewport(xscale = c(0, 20), yscale = c(-10, 10)))
		  grid.text(top.label, unit(10, "native"), unit(9, "native"))
		  grid.text(bottom.label, unit(10, "native"), unit(-9, "native"))
		  
		  popViewport(2)
		  
		}
		
		p <- grid.grabExpr(headTailPlot())
		
	  }
	  
	  
	"""		
		
		
#	with pd.option_context('display.max_rows', None, 'display.max_columns', None):
#		print(similarity_score)
		
	if output_list:
		return similarity_score, reverse_similarity_score, alignment
		# Unimplemented R code
		"""
		return(list(similarity.score = similarity_score, 
			alignment = alignment,
			plot = p))
		"""
	else:
		return similarity_score, reverse_similarity_score

	# simscore <- as.vector((u %*% v)^2 / (sum(u^2) * sum(v^2)))   # cos squared

def normalize(row, max_val):
	#http://jonathansoma.com/lede/foundations/classes/pandas%20columns%20and%20functions/apply-a-function-to-every-row-in-a-pandas-dataframe/
	return (row["intensity"]/float(max_val))*100.0	

def create_array(intensities, mz):
	return np.column_stack((mz,intensities))






#!/usr/bin/env python
#
#  constants.py
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

import logging

API_BASE = 'https://pubchem.ncbi.nlm.nih.gov/rest/pug'

log = logging.getLogger('pubchempy')
log.addHandler(logging.NullHandler())

text_types = str, bytes

ELEMENTS = {
		1: 'H',
		2: 'He',
		3: 'Li',
		4: 'Be',
		5: 'B',
		6: 'C',
		7: 'N',
		8: 'O',
		9: 'F',
		10: 'Ne',
		11: 'Na',
		12: 'Mg',
		13: 'Al',
		14: 'Si',
		15: 'P',
		16: 'S',
		17: 'Cl',
		18: 'Ar',
		19: 'K',
		20: 'Ca',
		21: 'Sc',
		22: 'Ti',
		23: 'V',
		24: 'Cr',
		25: 'Mn',
		26: 'Fe',
		27: 'Co',
		28: 'Ni',
		29: 'Cu',
		30: 'Zn',
		31: 'Ga',
		32: 'Ge',
		33: 'As',
		34: 'Se',
		35: 'Br',
		36: 'Kr',
		37: 'Rb',
		38: 'Sr',
		39: 'Y',
		40: 'Zr',
		41: 'Nb',
		42: 'Mo',
		43: 'Tc',
		44: 'Ru',
		45: 'Rh',
		46: 'Pd',
		47: 'Ag',
		48: 'Cd',
		49: 'In',
		50: 'Sn',
		51: 'Sb',
		52: 'Te',
		53: 'I',
		54: 'Xe',
		55: 'Cs',
		56: 'Ba',
		57: 'La',
		58: 'Ce',
		59: 'Pr',
		60: 'Nd',
		61: 'Pm',
		62: 'Sm',
		63: 'Eu',
		64: 'Gd',
		65: 'Tb',
		66: 'Dy',
		67: 'Ho',
		68: 'Er',
		69: 'Tm',
		70: 'Yb',
		71: 'Lu',
		72: 'Hf',
		73: 'Ta',
		74: 'W',
		75: 'Re',
		76: 'Os',
		77: 'Ir',
		78: 'Pt',
		79: 'Au',
		80: 'Hg',
		81: 'Tl',
		82: 'Pb',
		83: 'Bi',
		84: 'Po',
		85: 'At',
		86: 'Rn',
		87: 'Fr',
		88: 'Ra',
		89: 'Ac',
		90: 'Th',
		91: 'Pa',
		92: 'U',
		93: 'Np',
		94: 'Pu',
		95: 'Am',
		96: 'Cm',
		97: 'Bk',
		98: 'Cf',
		99: 'Es',
		100: 'Fm',
		101: 'Md',
		102: 'No',
		103: 'Lr',
		104: 'Rf',
		105: 'Db',
		106: 'Sg',
		107: 'Bh',
		108: 'Hs',
		109: 'Mt',
		110: 'Ds',
		111: 'Rg',
		112: 'Cp',
		113: 'ut',
		114: 'uq',
		115: 'up',
		116: 'uh',
		117: 'us',
		118: 'uo',
		}

# Allows properties to optionally be specified as underscore_separated, consistent with Compound attributes
PROPERTY_MAP = {
		'molecular_formula': 'MolecularFormula',
		'molecular_weight': 'MolecularWeight',
		'canonical_smiles': 'CanonicalSMILES',
		'isomeric_smiles': 'IsomericSMILES',
		'inchi': 'InChI',
		'inchikey': 'InChIKey',
		'iupac_name': 'IUPACName',
		'xlogp': 'XLogP',
		'exact_mass': 'ExactMass',
		'monoisotopic_mass': 'MonoisotopicMass',
		'tpsa': 'TPSA',
		'complexity': 'Complexity',
		'charge': 'Charge',
		'h_bond_donor_count': 'HBondDonorCount',
		'h_bond_acceptor_count': 'HBondAcceptorCount',
		'rotatable_bond_count': 'RotatableBondCount',
		'heavy_atom_count': 'HeavyAtomCount',
		'isotope_atom_count': 'IsotopeAtomCount',
		'atom_stereo_count': 'AtomStereoCount',
		'defined_atom_stereo_count': 'DefinedAtomStereoCount',
		'undefined_atom_stereo_count': 'UndefinedAtomStereoCount',
		'bond_stereo_count': 'BondStereoCount',
		'defined_bond_stereo_count': 'DefinedBondStereoCount',
		'undefined_bond_stereo_count': 'UndefinedBondStereoCount',
		'covalent_unit_count': 'CovalentUnitCount',
		'volume_3d': 'Volume3D',
		'conformer_rmsd_3d': 'ConformerModelRMSD3D',
		'conformer_model_rmsd_3d': 'ConformerModelRMSD3D',
		'x_steric_quadrupole_3d': 'XStericQuadrupole3D',
		'y_steric_quadrupole_3d': 'YStericQuadrupole3D',
		'z_steric_quadrupole_3d': 'ZStericQuadrupole3D',
		'feature_count_3d': 'FeatureCount3D',
		'feature_acceptor_count_3d': 'FeatureAcceptorCount3D',
		'feature_donor_count_3d': 'FeatureDonorCount3D',
		'feature_anion_count_3d': 'FeatureAnionCount3D',
		'feature_cation_count_3d': 'FeatureCationCount3D',
		'feature_ring_count_3d': 'FeatureRingCount3D',
		'feature_hydrophobe_count_3d': 'FeatureHydrophobeCount3D',
		'effective_rotor_count_3d': 'EffectiveRotorCount3D',
		'conformer_count_3d': 'ConformerCount3D',
		}


class CoordinateType:
	TWO_D = 1
	THREE_D = 2
	SUBMITTED = 3
	EXPERIMENTAL = 4
	COMPUTED = 5
	STANDARDIZED = 6
	AUGMENTED = 7
	ALIGNED = 8
	COMPACT = 9
	UNITS_ANGSTROMS = 10
	UNITS_NANOMETERS = 11
	UNITS_PIXEL = 12
	UNITS_POINTS = 13
	UNITS_STDBONDS = 14
	UNITS_UNKNOWN = 255


class ProjectCategory:
	MLSCN = 1
	MPLCN = 2
	MLSCN_AP = 3
	MPLCN_AP = 4
	JOURNAL_ARTICLE = 5
	ASSAY_VENDOR = 6
	LITERATURE_EXTRACTED = 7
	LITERATURE_AUTHOR = 8
	LITERATURE_PUBLISHER = 9
	RNAIGI = 10
	OTHER = 255


# The following from periodictable
# public domain data
# Author: Paul Kienzle

avogadro_number = 6.02214179e23  # (30) mol-1
plancks_constant = 4.13566733e-15  # (10) eV s
electron_volt = 1.602176487e-19  # (40) J / eV
speed_of_light = 299792458  # m/s (exact)
electron_radius = 2.8179402894e-15  # (58) m

# From NIST Reference on Constants, Units, and Uncertainty
#   http://physics.nist.gov/cuu/index.html
# neutron mass = 1.008 664 915 97(43) u
# atomic mass constant m_u = 1.660 538 782(83) x 10-27 kg
neutron_mass = 1.00866491597  # u
atomic_mass_constant = 1.660538782e-27  # (83) kg / u



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  !/usr/bin/env python
#   -*- coding: utf-8 -*-
#
#  property_format.py
"""Format Physical Properties for Chemicals"""

#
#  Copyright (c) 2019 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import re
from decimal import Decimal

def degC(string):
	return string.replace(" deg C", "°C").replace("deg C", "°C").replace(" DEG C", "°C").replace("DEG C", "°C")


def equals(string):
	return string.replace("= ", " = ").replace("  = ", " = ")


def scientific(string):
	"""
	TODO: Finish
	:param string:
	:return:
	"""
	import re
	try:
		magnitude = re.findall("X10.[0-9]+", string)[0].replace("X10", '').replace("-", "−")
	except IndexError:  # no scientific notation to format
		return string
	# print(magnitude)
	return re.sub("X10.[0-9]+", f"×10<sup>{magnitude}</sup>", string)


def uscg1999(string):
	return string.replace("(USCG, 1999)", '')


def trailspace(string):
	return string.rstrip(" ")


def f2c(string):
	try:
		temperature = re.findall(r"\d*\.?\d+ *° *F", string)[0].replace("F", '').replace("°", '').replace(" ", '')
	except IndexError:
		return string
	
	# Convert to Censius and strip trailing 0s and decimal points
	temperature = str((Decimal(temperature) - 32) * (Decimal(5) / Decimal(9)))
	if "." in temperature:
		temperature = temperature.rstrip("0").rstrip(".")
	return re.sub(r"\d*\.?\d+ *° *F", f"{temperature}°C", string)


def property_format(string):
	return trailspace(f2c(uscg1999(scientific(degC(equals(string))))))



#  !/usr/bin/env python
#
#  assay.py
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
import json

# this package
from .utils import request


class Assay:
	
	@classmethod
	def from_aid(cls, aid):
		"""
		Retrieve the Assay record for the specified AID.

		:param int aid: The PubChem Assay Identifier (AID).
		"""
		
		record = json.loads(request(aid, 'aid', 'assay', 'description').content.decode())['PC_AssayContainer'][0]
		return cls(record)
	
	def __init__(self, record):
		"""
		:param record: A dictionary containing the full Assay record that all other properties are obtained from.
		:type record: dict
		"""
		self.record = record
	
	def __repr__(self):
		return 'Assay(%s)' % self.aid if self.aid else 'Assay()'
	
	def __eq__(self, other):
		return isinstance(other, type(self)) and self.record == other.record
	
	def to_dict(self, properties=None):
		"""
		Return a dictionary containing Assay data.

		If the properties parameter is not specified, everything is included.

		:param properties: (optional) A list of the desired properties.
		"""
		
		if not properties:
			properties = [p for p in dir(Assay) if isinstance(getattr(Assay, p), property)]
		return {p: getattr(self, p) for p in properties}
	
	@property
	def aid(self):
		"""The PubChem Substance Identifier (SID)."""
		return self.record['assay']['descr']['aid']['id']
	
	@property
	def name(self):
		"""The short assay name, used for display purposes."""
		return self.record['assay']['descr']['name']
	
	@property
	def description(self):
		"""Description"""
		return self.record['assay']['descr']['description']
	
	@property
	def project_category(self):
		"""A category to distinguish projects funded through MLSCN, MLPCN or from literature.

		Possible values include mlscn, mlpcn, mlscn-ap, mlpcn-ap, literature-extracted, literature-author,
		literature-publisher, rnaigi.
		"""
		if 'project_category' in self.record['assay']['descr']:
			return self.record['assay']['descr']['project_category']
	
	@property
	def comments(self):
		"""Comments and additional information."""
		return [comment for comment in self.record['assay']['descr']['comment'] if comment]
	
	@property
	def results(self):
		"""A list of dictionaries containing details of the results from this Assay."""
		return self.record['assay']['descr']['results']
	
	@property
	def target(self):
		"""A list of dictionaries containing details of the Assay targets."""
		if 'target' in self.record['assay']['descr']:
			return self.record['assay']['descr']['target']
		else:
			return [{}]
	
	@property
	def revision(self):
		"""Revision identifier for textual description."""
		return self.record['assay']['descr']['revision']
	
	@property
	def aid_version(self):
		"""Incremented when the original depositor updates the record."""
		return self.record['assay']['descr']['aid']['version']



#  !/usr/bin/env python
#
#  lookup.py
"""
Lookup properties for compound by name or CAS number
Uses data from PubChem and toxnet
"""
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

from chemistry_tools.assay import Assay
from chemistry_tools.compound import Compound, compounds_to_frame
from chemistry_tools.substance import Substance, substances_to_frame
from chemistry_tools.utils import get_json


def get_compounds(identifier, namespace='cid', searchtype=None, as_dataframe=False, **kwargs):
	"""
	Retrieve the specified compound records from PubChem.

	:param identifier: The compound identifier to use as a search query.
	:param namespace: (optional) The identifier type, one of cid, name, smiles, sdf, inchi, inchikey or formula.
	:param searchtype: (optional) The advanced search type, one of substructure, superstructure or similarity.
	:param as_dataframe: (optional) Automatically extract the :class:`~pubchempy.Compound` properties into a pandas
						 :class:`~pandas.DataFrame` and return that.
	"""
	
	results = get_json(identifier, namespace, searchtype=searchtype, **kwargs)
	compounds = [Compound(r) for r in results['PC_Compounds']] if results else []
	if as_dataframe:
		return compounds_to_frame(compounds)
	return compounds


def get_assays(identifier, namespace='aid', **kwargs):
	"""Retrieve the specified assay records from PubChem.

	:param identifier: The assay identifier to use as a search query.
	:param namespace: (optional) The identifier type.
	"""
	results = get_json(identifier, namespace, 'assay', 'description', **kwargs)
	return [Assay(r) for r in results['PC_AssayContainer']] if results else []


def get_substances(identifier, namespace='sid', as_dataframe=False, **kwargs):
	"""Retrieve the specified substance records from PubChem.

	:param identifier: The substance identifier to use as a search query.
	:param namespace: (optional) The identifier type, one of sid, name or sourceid/<source name>.
	:param as_dataframe: (optional) Automatically extract the :class:`~pubchempy.Substance` properties into a pandas
						 :class:`~pandas.DataFrame` and return that.
	"""
	results = get_json(identifier, namespace, 'substance', **kwargs)
	substances = [Substance(r) for r in results['PC_Substances']] if results else []
	if as_dataframe:
		return substances_to_frame(substances)
	return substances





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  toxnet.py
"""Read data from National Library of Medicine TOXNET"""
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

from bs4 import BeautifulSoup
import requests
from decimal import Decimal
from .property_format import *

def toxnet(cas):
	try:
		base_url = "https://toxnet.nlm.nih.gov"
		originURL = f"{base_url}/cgi-bin/sis/search2/r?dbs+hsdb:@term+@rn+@rel+{cas}"
		origin_page = requests.get(originURL)
		origin_soup = BeautifulSoup(origin_page.text, "html.parser")
		#print(originURL)
		#print(origin_soup.find("a", {"id": "anch_103"}))
		#print(origin_soup.find("input", {"name": "dfield"}))
		dataURL = origin_soup.find("input", {"name": "dfield"}).find_next_sibling("a")["href"][:-4]+"cpp"
		#print(dataURL)
		data_page = requests.get(base_url + dataURL)
		data_soup = BeautifulSoup(data_page.text, "html.parser")
	except AttributeError:
		raise ValueError(f"No Record was found for {cas}")
	
	physical_properties = {}
	
	for property in data_soup.findAll("h3"):
		property_name = str(property).replace("<h3>",'').replace(":</h3>",'')
		property_value_and_unit = property.nextSibling.replace("\n",'')
		
		if property_name in ["Molecular Formula", "Molecular Weight"]:
			continue
		
		if property_name == "Solubilities":
			property_name = "Solubility"
		
		#elif property_name == "Vapor Density":
		#	physical_properties[property_name] = property_value_and_unit.replace(" (Air= 1)",'')
		#else:
		#	physical_properties[property_name] = property_format(property_value_and_unit)
		
		physical_properties[property_name] = {}
		
		# Parse values
		# Temperatures
		if property_name in ["Boiling Point","Melting Point"]:
			property_value = property_value_and_unit.split(" deg C")[0]
			try:
				physical_properties[property_name]["Value"] = Decimal(property_value)
			except:
				physical_properties[property_name]["Value"] = property_format(property_value)
			physical_properties[property_name]["Unit"] = "°C"
			physical_properties[property_name]["Description"] = None
	
		# Strings
		elif property_name in ["Color/Form", "Odor", "Other Chemical/Physical Properties", "Solubility", "Spectral Properties"]:
			physical_properties[property_name]["Value"] = property_format(property_value_and_unit)
			physical_properties[property_name]["Unit"] = None
			physical_properties[property_name]["Description"] = None
		
		# Custom Units &c.
		elif property_name == "Density/Specific Gravity":
			try:
				physical_properties[property_name]["Value"] = Decimal(property_value_and_unit)
			except:
				physical_properties[property_name]["Value"] = property_format(property_value_and_unit)
			physical_properties[property_name]["Unit"] = "kg/m<sup>3</sup>"
			physical_properties[property_name]["Description"] = None
		elif property_name == "Vapor Density":
			#property_value = property_value_and_unit.split(" ")[0]
			physical_properties[property_name]["Value"] = property_format(property_value_and_unit.replace(" (Air= 1)",''))
			physical_properties[property_name]["Unit"] = "None"
			physical_properties[property_name]["Description"] = "Air=1"
			
		# TODO
		elif property_name == "Dissociation Constants":
			#property_value_list = property_value_and_unit.split(" ")
			#property_value = property_value_list[2]
			#physical_properties[property_name]["Value"] = property_value
			#physical_properties[property_name]["Unit"] = f"pKa @ {property_value_list[4]}°C"
			physical_properties[property_name]["Unit"] = None
			physical_properties[property_name]["Value"] = property_format(property_value_and_unit)
			physical_properties[property_name]["Description"] = None
		elif property_name == "Heat of Combustion":
			physical_properties[property_name]["Value"] = property_format(property_value_and_unit)
			physical_properties[property_name]["Unit"] = None
			physical_properties[property_name]["Description"] = None
			#TODO Extract J/KG value from string
		elif property_name == "Octanol/Water Partition Coefficient":
			#property_value = property_value_and_unit.split("log Kow = ")[1]
			#print(property_value)
			#print(property_value_and_unit.split("log Kow = "))
			## TODO Fix
			#physical_properties[property_name]["Value"] = property_value
			#physical_properties[property_name]["Unit"] = "log Kow"
			physical_properties[property_name]["Value"] = property_format(property_value_and_unit)
			physical_properties[property_name]["Unit"] = None
			physical_properties[property_name]["Description"] = None
		elif property_name == "Surface Tension":
			#property_value_list = property_value_and_unit.split(" ")
			#property_value = property_value_list[3]
			#print(property_value_list)
			#physical_properties[property_name]["Value"] = property_value
			#physical_properties[property_name]["Unit"] = f"N/M @ {property_value_list[6]}°C"
			physical_properties[property_name]["Value"] = property_format(property_value_and_unit)
			physical_properties[property_name]["Unit"] = None
			physical_properties[property_name]["Description"] = None
		elif property_name == "Vapor Pressure":
			#property_value_list = property_value_and_unit.split(" ")
			#property_value = property_value_list[0]
			#physical_properties[property_name]["Value"] = property_value
			#physical_properties[property_name]["Unit"] = f"mm Hg @ {property_value_list[4]}°C"
			physical_properties[property_name]["Value"] = property_format(property_value_and_unit)
			physical_properties[property_name]["Unit"] = None
			physical_properties[property_name]["Description"] = None
		#else:
		#	physical_properties[property_name] = property_value_and_unit
	

	return physical_properties

# Soup CAMEO Link from PubChem page if necessary



if __name__ == "__main__":
	import pprint
	pprint.pprint(toxnet("122-39-4"))
	#pprint.pprint(toxnet("85-98-3")) # No Record
	#pprint.pprint(toxnet("71-43-2"))




#  !/usr/bin/env python
#
#  errors.py
"""Error handling functions"""
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

import functools
import json
import warnings


class PubChemPyError(Exception):
	"""
	Base class for all PubChemPy exceptions.
	"""
	pass


class ResponseParseError(PubChemPyError):
	"""
	PubChem response is uninterpretable.
	"""
	pass


class PubChemHTTPError(PubChemPyError):
	"""
	Generic error class to handle all HTTP error codes.
	"""
	
	def __init__(self, e):
		self.code = e.status_code
		self.msg = e.reason
		try:
			self.msg += ': %s' % json.loads(e.content.decode())['Fault']['Details'][0]
		except (ValueError, IndexError, KeyError):
			pass
		if self.code == 400:
			raise BadRequestError(self.msg)
		elif self.code == 404:
			raise NotFoundError(self.msg)
		elif self.code == 405:
			raise MethodNotAllowedError(self.msg)
		elif self.code == 504:
			raise TimeoutError(self.msg)
		elif self.code == 501:
			raise UnimplementedError(self.msg)
		elif self.code == 500:
			raise ServerError(self.msg)
	
	def __str__(self):
		return repr(self.msg)


class BadRequestError(PubChemHTTPError):
	"""
	Request is improperly formed (syntax error in the URL, POST body, etc.).
	"""
	
	def __init__(self, msg='Request is improperly formed'):
		self.msg = msg


class NotFoundError(PubChemHTTPError):
	"""
	The input record was not found (e.g. invalid CID).
	"""
	
	def __init__(self, msg='The input record was not found'):
		self.msg = msg


class MethodNotAllowedError(PubChemHTTPError):
	"""
	Request not allowed (such as invalid MIME type in the HTTP Accept header).
	"""
	
	def __init__(self, msg='Request not allowed'):
		self.msg = msg


class TimeoutError(PubChemHTTPError):
	"""
	The request timed out, from server overload or too broad a request.

	See :ref:`Avoiding TimeoutError <avoiding_timeouterror>` for more information.
	"""
	
	def __init__(self, msg='The request timed out'):
		self.msg = msg


class UnimplementedError(PubChemHTTPError):
	"""The requested operation has not (yet) been implemented by the server."""
	
	def __init__(self, msg='The requested operation has not been implemented'):
		self.msg = msg


class ServerError(PubChemHTTPError):
	"""Some problem on the server side (such as a database server down, etc.)."""
	
	def __init__(self, msg='Some problem on the server side'):
		self.msg = msg


def deprecated(message=None):
	"""Decorator to mark functions as deprecated. A warning will be emitted when the function is used."""
	
	def deco(func):
		@functools.wraps(func)
		def wrapped(*args, **kwargs):
			warnings.warn(
					message or 'Call to deprecated function {}'.format(func.__name__),
					category=PubChemPyDeprecationWarning,
					stacklevel=2
					)
			return func(*args, **kwargs)
		
		return wrapped
	
	return deco


class PubChemPyDeprecationWarning(Warning):
	"""Warning category for deprecated features."""
	pass



#!/usr/bin/env python
#
#  compound.py
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
import json
from collections import Counter
from decimal import Decimal
from itertools import zip_longest

# this package
from .atom import Atom
from .bond import Bond
from .constants import CoordinateType, ELEMENTS, log
from .errors import ResponseParseError
from .property_format import *
from .toxnet import toxnet
from .utils import _parse_prop, get_full_json, get_json, memoized_property, request


class Compound:
	"""Corresponds to a single record from the PubChem Compound database.

	The PubChem Compound database is constructed from the Substance database using a standardization and deduplication
	process. Each Compound is uniquely identified by a CID.
	"""
	
	def __init__(self, record):
		"""Initialize with a record dict from the PubChem PUG REST service.

		For most users, the ``from_cid()`` class method is probably a better way of creating Compounds.

		:param dict record: A compound record returned by the PubChem PUG REST service.
		"""
		self._record = None
		self._atoms = {}
		self._bonds = {}
		self.record = record
		self.hazards = []  # COSHH Hazards
		self.CAS = ''
		self._physical_properties = {}  # Physical Properties (dictionary)
		self._full_record = get_full_json(self.cid)
		self._parse_full_record()
	
	@property
	def record(self):
		"""The raw compound record returned by the PubChem PUG REST service."""
		return self._record
	
	@property
	def full_record(self):
		return self._full_record
	
	@record.setter
	def record(self, record):
		self._record = record
		log.debug('Created %s' % self)
		self._setup_atoms()
		self._setup_bonds()
		self._full_record = get_full_json(self.cid)
	
	def _parse_full_record(self):
		try:
			for record in self.full_record["Record"]["Section"]:
				if record["TOCHeading"] == "Chemical Safety":
					try:
						for hazard in record["Information"][0]["Value"]["StringWithMarkup"][0]["Markup"]:
							self.hazards.append(hazard["Extra"])
					except KeyError:
						pass
				
				elif record["TOCHeading"] == "Chemical and Physical Properties":
					for section in record["Section"]:
						if section["TOCHeading"] == "Computed Properties":
							for physical_property in section["Section"]:
								name = physical_property["TOCHeading"]
								self._physical_properties[name] = {}
								self._physical_properties[name]["Description"] = physical_property[
									"Description"]
								
								try:
									self._physical_properties[name]["Value"] = Decimal(
											str(physical_property["Information"][0]["Value"]["Number"][0]))
								except KeyError:
									value = physical_property["Information"][0]["Value"]["StringWithMarkup"][0][
										"String"]
									if value == "Yes":
										value = True
									elif value == "No":
										value = False
									
									self._physical_properties[name]["Value"] = value
								
								try:
									self._physical_properties[name]["Unit"] = \
										physical_property["Information"][0]["Value"]["Unit"]
								except KeyError:
									self._physical_properties[name]["Unit"] = None
						
						elif section["TOCHeading"] == "Experimental Properties":
							# First Try TOXNET
							try:
								self._physical_properties = {**self._physical_properties, **toxnet(self.CAS)}
							except ValueError:  # Not Found in TOXNET
								# import traceback
								# traceback.print_exc()
								pass
							for physical_property in section["Section"]:
								name = physical_property["TOCHeading"]
								
								if name == "Other Experimental Properties":
									name = "Other Chemical/Physical Properties"
								elif name == "LogKoa":
									name = "Octanol/Water Partition Coefficient"
								elif name == "Density":
									name = "Density/Specific Gravity"
								
								if (name in self._physical_properties) or (
										name in ["Molecular Formula", "Molecular Weight", "Physical Description"]):
									continue
								
								self._physical_properties[name] = {}
								self._physical_properties[name]["Description"] = physical_property[
									"Description"]
								
								# TODO: Skip Physical Description, :
								try:
									self._physical_properties[name]["Value"] = Decimal(
											str(physical_property["Information"][0]["Value"]["Number"][0]))
								except KeyError:
									value = property_format(
											physical_property["Information"][0]["Value"]["StringWithMarkup"][0][
												"String"])
									if value == "Yes":
										value = True
									elif value == "No":
										value = False
									
									self._physical_properties[name]["Value"] = value
								
								try:
									self._physical_properties[name]["Unit"] = \
										physical_property["Information"][0]["Value"]["Unit"]
								except KeyError:
									self._physical_properties[name]["Unit"] = None
							# import pprint
							# pprint.pprint(self._physical_properties)
				
				elif record["TOCHeading"] == "Names and Identifiers":
					for section in record["Section"]:
						if section["TOCHeading"] == "Record Description":
							for description in section["Information"]:
								try:
									if description["Description"] == "Physical Description":
										self.physical_description = description["Value"]["StringWithMarkup"][0]
									
									elif description["Description"] == "Ontology Summary":
										self.ontology = description["Value"]["StringWithMarkup"][0]
									
									elif description["Description"] == "Metabolite Description":
										self.metabolite = description["Value"]["StringWithMarkup"][0]
								except KeyError:
									pass
						
						elif section["TOCHeading"] == "Computed Descriptors":
							for subsection in section["Section"]:
								if subsection["TOCHeading"] == "IUPAC Name":
									self.IUPAC = subsection["Information"][0]["Value"]["StringWithMarkup"][0]
						
						elif section["TOCHeading"] == "Molecular Formula":
							self.formula = section["Information"][0]["Value"]["StringWithMarkup"][0]["String"]
						
						elif section["TOCHeading"] == "Other Identifiers":
							for subsection in section["Section"]:
								if subsection["TOCHeading"] == "CAS":
									self.CAS = subsection["Information"][0]["Value"]["StringWithMarkup"][0]["String"]
						
						elif section["TOCHeading"] == "Synonyms":
							for subsection in section["Section"]:
								if subsection["TOCHeading"] == "Depositor-Supplied Synonyms":
									self.synonyms = []
									existing_names = [x.lower().replace("-", '') for x in
													  [self.CAS, self.IUPAC["String"]]]
									for synonym in subsection["Information"][0]["Value"]["StringWithMarkup"]:
										if synonym["String"].lower().replace("-", '') not in existing_names:
											self.synonyms.append(synonym["String"])
											if len(self.synonyms) >= 10:
												break
		except KeyError:
			pass
	
	def _setup_atoms(self):
		"""Derive Atom objects from the record."""
		# Delete existing atoms
		self._atoms = {}
		# Create atoms
		aids = self.record['atoms']['aid']
		elements = self.record['atoms']['element']
		if not len(aids) == len(elements):
			raise ResponseParseError('Error parsing atom elements')
		for aid, element in zip(aids, elements):
			self._atoms[aid] = Atom(aid=aid, number=element)
		# Add coordinates
		if 'coords' in self.record:
			coord_ids = self.record['coords'][0]['aid']
			xs = self.record['coords'][0]['conformers'][0]['x']
			ys = self.record['coords'][0]['conformers'][0]['y']
			zs = self.record['coords'][0]['conformers'][0].get('z', [])
			if not len(coord_ids) == len(xs) == len(ys) == len(self._atoms) or (zs and not len(zs) == len(coord_ids)):
				raise ResponseParseError('Error parsing atom coordinates')
			for aid, x, y, z in zip_longest(coord_ids, xs, ys, zs):
				self._atoms[aid].set_coordinates(x, y, z)
		# Add charges
		if 'charge' in self.record['atoms']:
			for charge in self.record['atoms']['charge']:
				self._atoms[charge['aid']].charge = charge['value']
	
	def _setup_bonds(self):
		"""Derive Bond objects from the record."""
		self._bonds = {}
		if 'bonds' not in self.record:
			return
		# Create bonds
		aid1s = self.record['bonds']['aid1']
		aid2s = self.record['bonds']['aid2']
		orders = self.record['bonds']['order']
		if not len(aid1s) == len(aid2s) == len(orders):
			raise ResponseParseError('Error parsing bonds')
		for aid1, aid2, order in zip(aid1s, aid2s, orders):
			self._bonds[frozenset((aid1, aid2))] = Bond(aid1=aid1, aid2=aid2, order=order)
		# Add styles
		if 'coords' in self.record and 'style' in self.record['coords'][0]['conformers'][0]:
			aid1s = self.record['coords'][0]['conformers'][0]['style']['aid1']
			aid2s = self.record['coords'][0]['conformers'][0]['style']['aid2']
			styles = self.record['coords'][0]['conformers'][0]['style']['annotation']
			for aid1, aid2, style in zip(aid1s, aid2s, styles):
				self._bonds[frozenset((aid1, aid2))].style = style
	
	@classmethod
	def from_cid(cls, cid, **kwargs):
		"""Retrieve the Compound record for the specified CID.

		Usage::

			c = Compound.from_cid(6819)

		:param int cid: The PubChem Compound Identifier (CID).
		"""
		record = json.loads(request(cid, **kwargs).content.decode())['PC_Compounds'][0]
		return cls(record)
	
	def __repr__(self):
		return 'Compound(%s)' % self.cid if self.cid else 'Compound()'
	
	def __eq__(self, other):
		return isinstance(other, type(self)) and self.record == other.record
	
	def to_dict(self, properties=None):
		"""Return a dictionary containing Compound data. Optionally specify a list of the desired properties.

		synonyms, aids and sids are not included unless explicitly specified using the properties parameter. This is
		because they each require an extra request.
		"""
		if not properties:
			skip = {'aids', 'sids', 'synonyms'}
			properties = [p for p in dir(Compound) if isinstance(getattr(Compound, p), property) and p not in skip]
		return {p: [i.to_dict() for i in getattr(self, p)] if p in {'atoms', 'bonds'} else getattr(self, p) for p in
				properties}
	
	def to_series(self, properties=None):
		"""Return a pandas :class:`~pandas.Series` containing Compound data. Optionally specify a list of the desired
		properties.

		synonyms, aids and sids are not included unless explicitly specified using the properties parameter. This is
		because they each require an extra request.
		"""
		import pandas as pd
		return pd.Series(self.to_dict(properties))
	
	@staticmethod
	def format_string(stringwithmarkup):
		string = list(stringwithmarkup["String"])
		try:
			markup_list = stringwithmarkup["Markup"]
		except KeyError:
			markup_list = []
		
		for markup in markup_list:
			style = None
			start = markup["Start"]
			end = markup["Length"] + start - 1
			if markup["Type"] == "Italics":
				style = "i"
			# handle Other formats
			
			if style is None:
				print(markup)
				continue
			
			string[start] = f"<{style}>{string[start]}"
			string[end] = f"{string[end]}</{style}>"
		
		string = ''.join(string)
		
		return string
	
	"""Structure"""
	
	@property
	def elements(self):
		"""List of element symbols for atoms in this Compound."""
		return [a.element for a in self.atoms]
	
	@property
	def atoms(self):
		"""List of :class:`Atoms <pubchempy.Atom>` in this Compound."""
		return sorted(self._atoms.values(), key=lambda x: x.aid)
	
	@property
	def bonds(self):
		"""List of :class:`Bonds <pubchempy.Bond>` between :class:`Atoms <pubchempy.Atom>` in this Compound."""
		return sorted(self._bonds.values(), key=lambda x: (x.aid1, x.aid2))
	
	@memoized_property
	def sids(self):
		"""Requires an extra request. Result is cached."""
		if self.cid:
			results = get_json(self.cid, operation='sids')
			return results['InformationList']['Information'][0]['SID'] if results else []
	
	@memoized_property
	def aids(self):
		"""Requires an extra request. Result is cached."""
		if self.cid:
			results = get_json(self.cid, operation='aids')
			return results['InformationList']['Information'][0]['AID'] if results else []
	
	@property
	def coordinate_type(self):
		if CoordinateType.TWO_D in self.record['coords'][0]['type']:
			return '2d'
		elif CoordinateType.THREE_D in self.record['coords'][0]['type']:
			return '3d'
	
	@property
	def molecular_formula(self):
		"""Molecular formula."""
		return _parse_prop({'label': 'Molecular Formula'}, self.record['props'])
	
	"""Identifiers"""
	
	@property
	def cid(self):
		"""The PubChem Compound Identifier (CID).

		.. note::

			When searching using a SMILES or InChI query that is not present in the PubChem Compound database, an
			automatically generated record may be returned that contains properties that have been calculated on the
			fly. These records will not have a CID property.
		"""
		if 'id' in self.record and 'id' in self.record['id'] and 'cid' in self.record['id']['id']:
			return self.record['id']['id']['cid']
	
	@property
	def smiles(self):
		"""Canonical SMILES, with no stereochemistry information."""
		return _parse_prop({'label': 'SMILES', 'name': 'Canonical'}, self.record['props'])
	
	@property
	def canonical_smiles(self):
		"""Canonical SMILES, with no stereochemistry information."""
		return self.smiles
	
	@property
	def isomeric_smiles(self):
		"""Isomeric SMILES."""
		return _parse_prop({'label': 'SMILES', 'name': 'Isomeric'}, self.record['props'])
	
	@property
	def inchi(self):
		"""InChI string."""
		return _parse_prop({'label': 'InChI', 'name': 'Standard'}, self.record['props'])
	
	@property
	def inchikey(self):
		"""InChIKey."""
		return _parse_prop({'label': 'InChIKey', 'name': 'Standard'}, self.record['props'])
	
	@property
	def iupac_name(self):
		"""Preferred IUPAC name."""
		# Note: Allowed, CAS-like Style, Preferred, Systematic, Traditional are available in full record
		return _parse_prop({'label': 'IUPAC Name', 'name': 'Preferred'}, self.record['props'])
	
	@property
	def systematic_name(self):
		"""Systematic IUPAC name."""
		# Note: Allowed, CAS-like Style, Preferred, Systematic, Traditional are available in full record
		return _parse_prop({'label': 'IUPAC Name', 'name': 'Systematic'}, self.record['props'])
	
	@memoized_property
	def hill_formula(self):
		element_count = Counter(self.elements)
		hill = []
		
		alphabet = sorted([ELEMENTS[x] for x in ELEMENTS])
		
		if "C" in element_count:
			hill.append("C")
			alphabet.remove("C")
			count = element_count["C"]
			if count > 1:
				hill.append(f"<sub>{count}</sub>")
			if "H" in element_count:
				hill.append("H")
				alphabet.remove("H")
				count = element_count["H"]
				if count > 1:
					hill.append(f"<sub>{count}</sub>")
		
		for element in alphabet:
			if element in element_count:
				hill.append(element)
				count = element_count[element]
				if count > 1:
					hill.append(f"<sub>{count}</sub>")
		
		return ''.join(hill)
	
	@memoized_property
	def _synonyms(self):
		"""A ranked list of all the names associated with this Compound.

		Requires an extra request. Result is cached.
		"""
		if self.cid:
			results = get_json(self.cid, operation='synonyms')
			return results['InformationList']['Information'][0]['Synonym'] if results else []
	
	@property
	def fingerprint(self):
		"""Raw padded and hex-encoded fingerprint, as returned by the PUG REST API."""
		return _parse_prop({'implementation': 'E_SCREEN'}, self.record['props'])
	
	@property
	def cactvs_fingerprint(self):
		"""PubChem CACTVS fingerprint.

		Each bit in the fingerprint represents the presence or absence of one of 881 chemical substructures.

		More information at ftp://ftp.ncbi.nlm.nih.gov/pubchem/specifications/pubchem_fingerprints.txt
		"""
		# Skip first 4 bytes (contain length of fingerprint) and last 7 bits (padding) then re-pad to 881 bits
		return '{:020b}'.format(int(self.fingerprint[8:], 16))[:-7].zfill(881)
	
	"""Other"""
	
	@property
	def volume_3d(self):
		conf = self.record['coords'][0]['conformers'][0]
		if 'data' in conf:
			return _parse_prop({'label': 'Shape', 'name': 'Volume'}, conf['data'])
	
	@property
	def multipoles_3d(self):
		conf = self.record['coords'][0]['conformers'][0]
		if 'data' in conf:
			return _parse_prop({'label': 'Shape', 'name': 'Multipoles'}, conf['data'])
	
	@property
	def conformer_rmsd_3d(self):
		coords = self.record['coords'][0]
		if 'data' in coords:
			return _parse_prop({'label': 'Conformer', 'name': 'RMSD'}, coords['data'])
	
	@property
	def effective_rotor_count_3d(self):
		return _parse_prop({'label': 'Count', 'name': 'Effective Rotor'}, self.record['props'])
	
	@property
	def pharmacophore_features_3d(self):
		return _parse_prop({'label': 'Features', 'name': 'Pharmacophore'}, self.record['props'])
	
	@property
	def mmff94_partial_charges_3d(self):
		return _parse_prop({'label': 'Charge', 'name': 'MMFF94 Partial'}, self.record['props'])
	
	@property
	def mmff94_energy_3d(self):
		conf = self.record['coords'][0]['conformers'][0]
		if 'data' in conf:
			return _parse_prop({'label': 'Energy', 'name': 'MMFF94 NoEstat'}, conf['data'])
	
	@property
	def conformer_id_3d(self):
		conf = self.record['coords'][0]['conformers'][0]
		if 'data' in conf:
			return _parse_prop({'label': 'Conformer', 'name': 'ID'}, conf['data'])
	
	@property
	def shape_selfoverlap_3d(self):
		conf = self.record['coords'][0]['conformers'][0]
		if 'data' in conf:
			return _parse_prop({'label': 'Shape', 'name': 'Self Overlap'}, conf['data'])
	
	@property
	def feature_selfoverlap_3d(self):
		conf = self.record['coords'][0]['conformers'][0]
		if 'data' in conf:
			return _parse_prop({'label': 'Feature', 'name': 'Self Overlap'}, conf['data'])
	
	@property
	def shape_fingerprint_3d(self):
		conf = self.record['coords'][0]['conformers'][0]
		if 'data' in conf:
			return _parse_prop({'label': 'Fingerprint', 'name': 'Shape'}, conf['data'])
	
	def get_property_description(self, property):
		try:
			return self.get_property(property)["Description"]
		except KeyError:
			return
	
	def get_property_value(self, property):
		try:
			return self.get_property(property)["Value"]
		except KeyError:
			return "NotFound"
	
	def get_property_unit(self, property):
		try:
			return self.get_property(property)["Unit"]
		except KeyError:
			return
	
	def get_property(self, property):
		# TODO Error handling
		try:
			return self._physical_properties[property]
		except KeyError:
			return {"Value": "NotFound", "Unit": None, "Description": None}
	
	"""PubChem Computed Properties"""
	
	@property
	def molecular_mass(self):
		"""Molecular Mass."""
		return self.get_property_value("Molecular Weight")
	
	# return _parse_prop({'label': 'Molecular Weight'}, self.record['props'])
	
	@property
	def molecular_weight(self):
		"""Molecular Weight."""
		return self.molecular_mass
	
	@property
	def xlogp(self):
		"""XLogP."""
		return _parse_prop({'label': 'Log P'}, self.record['props'])
	
	@property
	def h_bond_donor_count(self):
		"""Hydrogen bond donor count."""
		return self.get_property_value("Hydrogen Bond Donor Count")
	
	# return _parse_prop({'implementation': 'E_NHDONORS'}, self.record['props'])
	
	@property
	def h_bond_acceptor_count(self):
		"""Hydrogen bond acceptor count."""
		return self.get_property_value("Hydrogen Bond Acceptor Count")
	
	# return _parse_prop({'implementation': 'E_NHACCEPTORS'}, self.record['props'])
	
	@property
	def rotatable_bond_count(self):
		"""Rotatable bond count."""
		return self.get_property_value("Rotatable Bond Count")
	
	# return _parse_prop({'implementation': 'E_NROTBONDS'}, self.record['props'])
	
	@property
	def exact_mass(self):
		"""Exact mass."""
		return self.get_property_value("Exact Mass")
	
	# return _parse_prop({'label': 'Mass', 'name': 'Exact'}, self.record['props'])
	
	@property
	def monoisotopic_mass(self):
		"""Monoisotopic mass."""
		return self.get_property_value("Monoisotopic Mass")
	
	# return _parse_prop({'label': 'Weight', 'name': 'MonoIsotopic'}, self.record['props'])
	
	@property
	def tpsa(self):
		"""Topological Polar Surface Area."""
		return self.get_property_value("Topological Polar Surface Area")
	
	# return _parse_prop({'implementation': 'E_TPSA'}, self.record['props'])
	
	@property
	def heavy_atom_count(self):
		"""Heavy atom count."""
		return self.get_property_value("Heavy Atom Count")
	
	# if 'count' in self.record and 'heavy_atom' in self.record['count']:
	#	return self.record['count']['heavy_atom']
	
	@property
	def charge(self):
		"""Formal charge on this Compound."""
		return self.get_property_value("Formal Charge")
	
	# return self.record['charge'] if 'charge' in self.record else 0
	
	@property
	def complexity(self):
		"""Complexity."""
		return self.get_property_value("Complexity")
	
	# return _parse_prop({'implementation': 'E_COMPLEXITY'}, self.record['props'])
	
	@property
	def isotope_atom_count(self):
		"""Isotope atom count."""
		return self.get_property_value("Isotope Atom Count")
	
	# if 'count' in self.record and 'isotope_atom' in self.record['count']:
	#	return self.record['count']['isotope_atom']
	
	@property
	def defined_atom_stereo_count(self):
		"""Defined atom stereocenter count."""
		return self.get_property_value("Defined Atom Stereocenter Count")
	
	# if 'count' in self.record and 'atom_chiral_def' in self.record['count']:
	#	return self.record['count']['atom_chiral_def']
	
	@property
	def undefined_atom_stereo_count(self):
		"""Undefined atom stereocenter count."""
		return self.get_property_value("Undefined Atom Stereocenter Count")
	
	# if 'count' in self.record and 'atom_chiral_undef' in self.record['count']:
	#	return self.record['count']['atom_chiral_undef']
	
	@property
	def defined_bond_stereo_count(self):
		"""Defined bond stereocenter count."""
		return self.get_property_value("Defined Bond Stereocenter Count")
	
	# if 'count' in self.record and 'bond_chiral_def' in self.record['count']:
	#	return self.record['count']['bond_chiral_def']
	
	@property
	def undefined_bond_stereo_count(self):
		"""Undefined bond stereocenter count."""
		return self.get_property_value("Undefined Bond Stereocenter Count")
	
	# if 'count' in self.record and 'bond_chiral_undef' in self.record['count']:
	#	return self.record['count']['bond_chiral_undef']
	
	@property
	def covalent_unit_count(self):
		"""Covalently-bonded unit count."""
		return self.get_property_value("Covalently-Bonded Unit Count")
	
	# if 'count' in self.record and 'covalent_unit' in self.record['count']:
	#	return self.record['count']['covalent_unit']
	
	@property
	def is_canonicalized(self):
		"""Compound Is Canonicalized"""
		return self.get_property_value("Compound Is Canonicalized")
	
	@property
	def atom_stereo_count(self):
		"""Atom stereocenter count."""
		if 'count' in self.record and 'atom_chiral' in self.record['count']:
			return self.record['count']['atom_chiral']
	
	@property
	def bond_stereo_count(self):
		"""Bond stereocenter count."""
		if 'count' in self.record and 'bond_chiral' in self.record['count']:
			return self.record['count']['bond_chiral']
	
	"""TOXNET / Experimental Properties"""
	
	@property
	def boiling_point(self):
		"""Boiling Point"""
		return self.get_property_value("Boiling Point")
	
	@property
	def color(self):
		"""Color/Form"""
		return self.get_property_value("Color/Form")
	
	@property
	def density(self):
		"""Density/Specific Gravity"""
		return self.get_property_value("Density/Specific Gravity")
	
	@property
	def specific_gravity(self):
		"""Density/Specific Gravity"""
		return self.get_property_value("Density/Specific Gravity")
	
	@property
	def dissociation_constant(self):
		"""Dissociation Constants"""
		return self.get_property_value("Dissociation Constants")
	
	@property
	def heat_combustion(self):
		"""Heat of Combustion"""
		return self.get_property_value("Heat of Combustion")
	
	@property
	def melting_point(self):
		"""Melting Point"""
		return self.get_property_value("Melting Point")
	
	@property
	def partition_coeff(self):
		"""Octanol/Water Partition Coefficient"""
		return self.get_property_value("Octanol/Water Partition Coefficient")
	
	@property
	def odor(self):
		"""Odor"""
		return self.get_property_value("Odor")
	
	@property
	def other_props(self):
		"""Other Chemical/Physical Properties"""
		return self.get_property_value("Other Chemical/Physical Properties")
	
	@property
	def solubility(self):
		"""Solubility"""
		return self.get_property_value("Solubility")
	
	@property
	def spectral_props(self):
		"""Spectral Properties"""
		return self.get_property_value("Spectral Properties")
	
	@property
	def surface_tension(self):
		"""Surface Tension"""
		return self.get_property_value("Surface Tension")
	
	@property
	def vapor_density(self):
		"""Vapor Density"""
		return self.get_property_value("Vapor Density")
	
	@property
	def vapor_pressure(self):
		"""Vapor Pressure"""
		return self.get_property_value("Vapor Pressure")


class CompoundIdType:
	#: Original Deposited Compound
	DEPOSITED = 0
	#: Standardized Form of the Deposited Compound
	STANDARDIZED = 1
	#: Component of the Standardized Form
	COMPONENT = 2
	#: Neutralized Form of the Standardized Form
	NEUTRALIZED = 3
	#: Deposited Mixture Component
	MIXTURE = 4
	#: Alternate Tautomer Form of the Standardized Form
	TAUTOMER = 5
	#: Ionized pKa Form of the Standardized Form
	IONIZED = 6
	#: Unspecified or Unknown Compound Type
	UNKNOWN = 255


def compounds_to_frame(compounds, properties=None):
	"""
	Construct a pandas :class:`~pandas.DataFrame` from a list of :class:`~pubchempy.Compound` objects.

	Optionally specify a list of the desired :class:`~pubchempy.Compound` properties.
	"""
	
	import pandas as pd
	if isinstance(compounds, Compound):
		compounds = [compounds]
	properties = set(properties) | {'cid'} if properties else None
	return pd.DataFrame.from_records([c.to_dict(properties) for c in compounds], index='cid')



#!/usr/bin/env python
#
#  bond.py
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

from .errors import deprecated


class BondType:
	SINGLE = 1
	DOUBLE = 2
	TRIPLE = 3
	QUADRUPLE = 4
	DATIVE = 5
	COMPLEX = 6
	IONIC = 7
	UNKNOWN = 255


class Bond:
	"""
	Class to represent a bond between two atoms in a :class:`~pubchempy.Compound`.
	"""
	
	def __init__(self, aid1, aid2, order=BondType.SINGLE, style=None):
		"""Initialize with begin and end atom IDs, bond order and bond style.

		:param int aid1: Begin atom ID.
		:param int aid2: End atom ID.
		:param int order: Bond order.
		"""
		self.aid1 = aid1
		# ID of the begin atom of this bond.
		self.aid2 = aid2
		# ID of the end atom of this bond.
		self.order = order
		# Bond order.
		self.style = style
		# Bond style annotation.
	
	def __repr__(self):
		return 'Bond({}, {}, {})'.format(self.aid1, self.aid2, self.order)
	
	def __eq__(self, other):
		return (
				isinstance(other, type(self)) and
				self.aid1 == other.aid1 and
				self.aid2 == other.aid2 and
				self.order == other.order and
				self.style == other.style
				)
	
	@deprecated('Dictionary style access to Bond attributes is deprecated')
	def __getitem__(self, prop):
		"""Allow dict-style access to attributes to ease transition from when bonds were dicts."""
		if prop in {'order', 'style'}:
			return getattr(self, prop)
		raise KeyError(prop)
	
	@deprecated('Dictionary style access to Bond attributes is deprecated')
	def __setitem__(self, prop, val):
		"""Allow dict-style setting of attributes to ease transition from when bonds were dicts."""
		setattr(self, prop, val)
	
	@deprecated('Dictionary style access to Atom attributes is deprecated')
	def __contains__(self, prop):
		"""Allow dict-style checking of attributes to ease transition from when bonds were dicts."""
		if prop in {'order', 'style'}:
			return getattr(self, prop) is not None
		return False
	
	@deprecated('Dictionary style access to Atom attributes is deprecated')
	def __delitem__(self, prop):
		"""Delete the property prop from the wrapped object."""
		if not hasattr(self.__wrapped, prop):
			raise KeyError(prop)
		delattr(self.__wrapped, prop)
	
	def to_dict(self):
		"""Return a dictionary containing Bond data."""
		data = {'aid1': self.aid1, 'aid2': self.aid2, 'order': self.order}
		if self.style is not None:
			data['style'] = self.style
		return data




#  !/usr/bin/env python
#
#  substance.py
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
import json

# this package
from .compound import Compound, CompoundIdType
from .utils import get_json, memoized_property, request


class Substance:
	"""
	Corresponds to a single record from the PubChem Substance database.

	The PubChem Substance database contains all chemical records deposited in PubChem in their most raw form, before
	any significant processing is applied. As a result, it contains duplicates, mixtures, and some records that don't
	make chemical sense. This means that Substance records contain fewer calculated properties, however they do have
	additional information about the original source that deposited the record.

	The PubChem Compound database is constructed from the Substance database using a standardization and deduplication
	process. Hence each Compound may be derived from a number of different Substances.
	"""
	
	@classmethod
	def from_sid(cls, sid):
		"""
		Retrieve the Substance record for the specified SID.

		:param int sid: The PubChem Substance Identifier (SID).
		"""
		
		# record = json.loads(request(sid, 'sid', 'substance').read().decode())['PC_Substances'][0]
		record = json.loads(request(sid, 'sid', 'substance').content.decode())['PC_Substances'][0]
		return cls(record)
	
	def __init__(self, record):
		"""
		:param record: A dictionary containing the full Substance record that all other properties are obtained from.
		:type record: dict
		"""

		self.record = record
	
	def __repr__(self):
		return 'Substance(%s)' % self.sid if self.sid else 'Substance()'
	
	def __eq__(self, other):
		return isinstance(other, type(self)) and self.record == other.record
	
	def to_dict(self, properties=None):
		"""Return a dictionary containing Substance data.

		If the properties parameter is not specified, everything except cids and aids is included. This is because the
		aids and cids properties each require an extra request to retrieve.

		:param properties: (optional) A list of the desired properties.
		"""
		if not properties:
			skip = {'deposited_compound', 'standardized_compound', 'cids', 'aids'}
			properties = [p for p in dir(Substance) if isinstance(getattr(Substance, p), property) and p not in skip]
		return {p: getattr(self, p) for p in properties}
	
	def to_series(self, properties=None):
		"""Return a pandas :class:`~pandas.Series` containing Substance data.

		If the properties parameter is not specified, everything except cids and aids is included. This is because the
		aids and cids properties each require an extra request to retrieve.

		:param properties: (optional) A list of the desired properties.
		"""
		import pandas as pd
		return pd.Series(self.to_dict(properties))
	
	@property
	def sid(self):
		"""The PubChem Substance Idenfitier (SID)."""
		return self.record['sid']['id']
	
	@property
	def synonyms(self):
		"""A ranked list of all the names associated with this Substance."""
		if 'synonyms' in self.record:
			return self.record['synonyms']
	
	@property
	def source_name(self):
		"""The name of the PubChem depositor that was the source of this Substance."""
		return self.record['source']['db']['name']
	
	@property
	def source_id(self):
		"""Unique ID for this Substance within those from the same PubChem depositor source."""
		return self.record['source']['db']['source_id']['str']
	
	@property
	def standardized_cid(self):
		"""The CID of the Compound that was produced when this Substance was standardized.

		May not exist if this Substance was not standardizable.
		"""
		for c in self.record['compound']:
			if c['id']['type'] == CompoundIdType.STANDARDIZED:
				return c['id']['id']['cid']
	
	@memoized_property
	def standardized_compound(self):
		"""Return the :class:`~pubchempy.Compound` that was produced when this Substance was standardized.

		Requires an extra request. Result is cached.
		"""
		for c in self.record['compound']:
			if c['id']['type'] == CompoundIdType.STANDARDIZED:
				return Compound.from_cid(c['id']['id']['cid'])
	
	@property
	def deposited_compound(self):
		"""Return a :class:`~pubchempy.Compound` produced from the unstandardized Substance record as deposited.

		The resulting :class:`~pubchempy.Compound` will not have a ``cid`` and will be missing most properties.
		"""
		for c in self.record['compound']:
			if c['id']['type'] == CompoundIdType.DEPOSITED:
				return Compound(c)
	
	@memoized_property
	def cids(self):
		"""A list of all CIDs for Compounds that were produced when this Substance was standardized.

		Requires an extra request. Result is cached."""
		results = get_json(self.sid, 'sid', 'substance', 'cids')
		return results['InformationList']['Information'][0]['CID'] if results else []
	
	@memoized_property
	def aids(self):
		"""A list of all AIDs for Assays associated with this Substance.

		Requires an extra request. Result is cached."""
		results = get_json(self.sid, 'sid', 'substance', 'aids')
		return results['InformationList']['Information'][0]['AID'] if results else []


def substances_to_frame(substances, properties=None):
	"""
	Construct a pandas :class:`~pandas.DataFrame` from a list of :class:`~pubchempy.Substance` objects.

	Optionally specify a list of the desired :class:`~pubchempy.Substance` properties.
	"""
	
	import pandas as pd
	if isinstance(substances, Substance):
		substances = [substances]
	properties = set(properties) | {'sid'} if properties else None
	return pd.DataFrame.from_records([s.to_dict(properties) for s in substances], index='sid')

# def add_columns_to_frame(dataframe, id_col, id_namespace, add_cols):
#     """"""
#     # Existing dataframe with some identifier column
#     # But consider what to do if the identifier column is an index?
#     # What about having the Compound/Substance object as a column?
